DAT fileName  byte "AS5055Demo120426a", 0
CON
{{ ****** Public Notes ******
  
  Written by: Duane Degn
  For: Kirk Fraser
  March 24, 2012

  See comments in child object.

  This demo only uses the 4-wire daisy-chain
  capability of the AS5055Spi120324a object.

  In daisy-chain mode the device at the end
  of the chain is considered device #0.

  Presently this object outputs data to a
  terminal window only when there is a change
  in the value to display.

  The display is formatted on the terminal window.
  With the new value overwriting the previous value.

  If you'd like to preserve a record of the values
  in the terminal window, change the value of
  "displayMode" in the first line of the Setup
  method to "_ScrollDisplay".
  
  This will cause the values to scroll down
  the terminal window leaving a record of the readings.

  
}}
{ ****** Private Notes ****** 
  Used servo tester as starting point.
  120316a Start writing software to read
  AS5055 magnetic encoders.
  17a Modify SPI driver to read 16 bits.
  Start with three wire read only mode.
  17b SPI object returns 32 bits sometimes.
  Why?
  18a Try again with MOSI pulled high.
  18a Appears to work a lot better.
  18b Clear the temp variable in the
  child object so return values are 16 bits
  not 32.
  18c Still 32 bits.  I had cleared the wrong
  variable.
  18d the parity bit is always zero.  I think
  I'm reading in one too many bits.
  18e Change the way child reads data in.
  18e Works much better than previous versions.
  18f Change output to 7-segment display.
  18g Change to expect twelve bits.  Instead of
  14 bits.
  18g Works really well with three wire SPI
  read only.
  LED display still needs some formatting help.
  18h Try to add second display mode to output
  raw values.
  Add a timer to change display modes.
  If output is between 40 and 50 degrees for two
  seconds and then moves to between 120 and 130 degrees
  within half a second and remains there for two seconds,
  the output will change to raw data mode.
  18k Add second PST object for 7-segment display.
  18k Read out jumps around with last digit.
  20a Try to get a more stable reading by averaging
  values.
  20a Works well except when near transition from
  4095 to 0.  The average near this transition
  can produce strange results.
  20b Try to come up with a way to smooth out
  transition.
  20c Seems to be working.
  21a Start adding daisy-chain capability.
  21b Simplify program.
  23b Daisy-chaining works!
  23d Add another buffer for averaging fast moves.
  23d didn't work correctly.  The displays froze up.
  23e Still didn't work.  AverageM equaled zero.
  23f Abandon moving average stuff.
  Work on cleaning up code.
  I don't think the second encoder averaging
  works yet.
  23g Strip code down to bare minimum for now.
  22bb works with two averaging buffers.
  23i Revert to 23e try to add moving average back in.
  23i Works.
  24a Clean up code, comment out 7-segment display code.
  
  Things to add.
  Check parity bit.
 
}
CON

  _CLKMODE = XTAL1 + PLL16X
  _CLKFREQ = 80_000_000

  _TwelveBits = %1111_1111_1111
  _TwelveBitsPlusOne = _TwelveBits + 1

  '' Set these pins to correspond with how the
  '' devices are connected to the Propeller chip.
  _IntPin = 8   ' input active low
  _MosiPin = 9  ' output
  _MisoPin = 10 ' input
  _SckPin = 11  ' output
  _SsPin = 12   ' output active low

  '' Baud rate used to communicate with the PC terminal window.
  _PstBaud = 57600

  ' Constants used with the small 7-segment displays
  {_SevenSegmentBaud = 9600 
  _DebugSevenSegment0TxPin = 14
  _DebugSevenSegment1TxPin = 15
  
  _SpecialCharacter = $77
  _DecimalPoint0 = %1
  _DecimalPoint1 = %10
  _DecimalPoint2 = %100
  _DecimalPointOff = 0  }

  '' The three constants (_SamplesToAverage, _SamplesToAverageMoving and
  '' _MovingTrigger) control the software filters.
  ''
  '' Small sample values will give faster position updates at the expense
  '' of less stable readings.
  ''
  '' _MovingTrigger is how large a change from one reading to the next
  '' is needed to trigger the smaller sample set used to compute the average.
  _SamplesToAverage = 64 '128
  _SamplesToAverageMoving = 16
  _MovingTrigger = 8            ' difference between two concecutive readings to trigger using small buffer to average
  _RolloverZone = 1000

  '' Set the constant "_ChainedDevices" to the number of devices that
  '' are to be monitored in daisy-chain mode. 
  _ChainedDevices = 1

  _ColumnWidth = 12   ' Terminal Window Formatting Constants
  _RowWidth = 4
  _DevicesPerRow = 4
  _DisplayStartY = 10
  _DisplayStartX = 4
  _TitleToReadingRows = 1
  _DisplayDigits = 4

  ' movingState enumeration
  #0, _NotMovingUpdatedBuffer, _Moving, _NotMovingFillingBuffer
    
  ' rolloverState enumeration
  #0, _OutOfZoneEmptyBuffer, _InZoneFillingBuffer, _InZoneUseBuffer, _OutOfZoneEmptyingBuffer

  ' localMovingFlag enumeration
  #0, _NotMovingThisLoop, _MovingThisLoop

  ' displayMode enumeration
  #0, _ScrollDisplay, _StationaryDisplay
  
VAR

  long spiCog, encoderAngle[_ChainedDevices], encoderAngle7[_ChainedDevices], previousAngle[_ChainedDevices]
  long angleDegrees, pseudoRealMultiplier, fullCircleDegrees
  long transitionMode, from1stLow[2], from1stHigh[2], from2ndLow[2], from2ndHigh[2] 
  long transitionInterval, transitionTime, holdPositionInterval
  long averagingBuffer[_SamplesToAverage * _ChainedDevices], bufferTotal[_ChainedDevices], average[_ChainedDevices]
  long bufferHead, rolloverBuffer[_SamplesToAverage * _ChainedDevices], rolloverCount[_ChainedDevices]
  long rolloverTotal[_ChainedDevices]
  long averagingBufferM[_SamplesToAverageMoving * _ChainedDevices], bufferTotalM[_ChainedDevices], averageM[_ChainedDevices]
  long bufferHeadM, rolloverBufferM[_SamplesToAverageMoving * _ChainedDevices], rolloverCountM[_ChainedDevices]
  long rolloverTotalM[_ChainedDevices], notMovingCount[_ChainedDevices]
  
  byte movingState[_ChainedDevices]
  byte rolloverState[_ChainedDevices], rolloverStateM[_ChainedDevices]
  byte displayX[_ChainedDevices], displayY[_ChainedDevices]
  byte titleX[_ChainedDevices], titleY[_ChainedDevices]
  byte colHeadingSize, dataDisplayOffset, displayMode
    
OBJ

  Debug : "Parallax Serial Terminal"
  'SevenSegment[2] : "Parallax Serial Terminal"  
  Spi : "AS5055Spi120324a"
  
PUB Setup | localIndex
'' First method executed by the Propeller.
'' This method initializes variables and
'' starts drivers.  

  displayMode := _StationaryDisplay  ' _ScrollDisplay
  colHeadingSize := strsize(@columnHeading)
  dataDisplayOffset := ((colHeadingSize + 1) - _DisplayDigits) / 2
  repeat localIndex from 0 to _ChainedDevices - 1
    rolloverState[localIndex] := _OutOfZoneEmptyBuffer
    rolloverStateM[localIndex] := _OutOfZoneEmptyBuffer
    movingState[localIndex] := _Moving
    titleY[localIndex] := _DisplayStartY + ((localIndex / _DevicesPerRow) * _RowWidth) 
    displayY[localIndex] := titleY[localIndex] + _TitleToReadingRows
    titleX[localIndex] := (localIndex // _DevicesPerRow) * _ColumnWidth
    displayX[localIndex] := titleX[localIndex] + dataDisplayOffset 
       
  {SevenSegment[0].StartRxTx(-1, _DebugSevenSegment0TxPin, 0, _SevenSegmentBaud)
  SevenSegment[1].StartRxTx(-1, _DebugSevenSegment1TxPin, 0, _SevenSegmentBaud)
  
  SevenSegment[0].Char("z")
  SevenSegment[0].Char(0)   ' Turn up brightness?
  SevenSegment[1].Char("z")
  SevenSegment[1].Char(0)   ' Turn up brightness?
  }
  
  Debug.Start(_PstBaud)
    
    
  waitcnt(clkfreq / 20 + cnt)
  spiCog := Spi.Start(_MosiPin, _MisoPin, _SckPin, _SsPin) - 1
  
  waitcnt(clkfreq * 3 + cnt) ' Time to open terminal window.                          
  DisplayHeading
  MainLoop     

PUB MainLoop | pinState, localIndex, errorFlag
'' The main loop of the program.

  Spi.Reset(_ChainedDevices)                            ' Reset all devices
  Spi.ReadChain(@encoderAngle, _ChainedDevices)         ' Output from devices lags one command,
                                                        ' so the returned value from this intial read command                                                        ' will be read while sending the next read command.
  repeat
    pinState := ina[_IntPin]
    if pinState == 0
      Spi.ReadChain(@encoderAngle, _ChainedDevices)
            
      if ++bufferHead == _SamplesToAverage
        bufferHead := 0  
      if ++bufferHeadM == _SamplesToAverageMoving
        bufferHeadM := 0
      errorFlag := 0
      repeat localIndex from 0 to _ChainedDevices - 1
        errorFlag += ProcessAngle(localIndex)
      if errorFlag > 0
        Debug.str(string(13, " ****** Error ****** "))
        Debug.str(string(13, " Reseting Devices "))
        Spi.Reset(_ChainedDevices)                      ' I'm not sure if this will clear the error.
                                                        ' I might need to add a separate "clear error" method                                                        
        Spi.ReadChain(@encoderAngle, _ChainedDevices)   ' Send read command so next read will have data to retrieve.
        
PRI ProcessAngle(localIndex) | parity, error, localMovingFlag
'' Parity bit is not tested or used.
'' localIndex is the ID of the device to process.
'' The device connected to the Prop's MISO pin is
'' device #0.
'' This method performs all the software filtering of the data.
       
  parity := encoderAngle[localIndex] & %1
  encoderAngle[localIndex] >>= 1
  error := encoderAngle[localIndex] & %1
  
  if error == 1 ' If there is an error, display the binary value of the data.
    Debug.str(string(13, " encoder #"))
    Debug.dec(localIndex)
    Debug.str(string(" returned an error. %")) 
    Debug.bin(encoderAngle[localIndex], 31)
    Debug.bin(parity, 1)
    result := 1                 ' Inform calling method of error.
  encoderAngle[localIndex] >>= 1
  encoderAngle[localIndex] &= _TwelveBits

  ' Compute averages, check to see if value is near rollover point and
  ' check to see if the encoder is moving.
  
  bufferTotal[localIndex] -= averagingBuffer[(localIndex * _SamplesToAverage) + bufferHead]
  averagingBuffer[(localIndex * _SamplesToAverage) + bufferHead] := encoderAngle[localIndex]
  bufferTotal[localIndex] += averagingBuffer[(localIndex * _SamplesToAverage) + bufferHead]
  average[localIndex] := bufferTotal[localIndex] / _SamplesToAverage

  bufferTotalM[localIndex] -= averagingBufferM[(localIndex * _SamplesToAverageMoving) + bufferHeadM]
  averagingBufferM[(localIndex * _SamplesToAverageMoving) + bufferHeadM] := encoderAngle[localIndex]
  bufferTotalM[localIndex] += averagingBufferM[(localIndex * _SamplesToAverageMoving) + bufferHeadM]
  averageM[localIndex] := bufferTotalM[localIndex] / _SamplesToAverageMoving

   
  if encoderAngle[localIndex] < _RolloverZone or encoderAngle[localIndex] > _TwelveBits - _RolloverZone
    rolloverTotal[localIndex] -= rolloverBuffer[(localIndex * _SamplesToAverage) + bufferHead]
    rolloverTotalM[localIndex] -= rolloverBufferM[(localIndex * _SamplesToAverageMoving) + bufferHeadM]
    if encoderAngle[localIndex] < _RolloverZone
      rolloverBuffer[(localIndex * _SamplesToAverage) + bufferHead] := encoderAngle[localIndex] + _TwelveBitsPlusOne
      rolloverBufferM[(localIndex * _SamplesToAverageMoving) + bufferHeadM] := encoderAngle[localIndex] + _TwelveBitsPlusOne
    else
      rolloverBuffer[(localIndex * _SamplesToAverage) + bufferHead] := encoderAngle[localIndex]
      rolloverBufferM[(localIndex * _SamplesToAverageMoving) + bufferHeadM] := encoderAngle[localIndex]
    rolloverTotal[localIndex] += rolloverBuffer[(localIndex * _SamplesToAverage) + bufferHead]
    rolloverTotalM[localIndex] += rolloverBufferM[(localIndex * _SamplesToAverageMoving) + bufferHeadM]  
    if rolloverState[localIndex] == _OutOfZoneEmptyBuffer
      rolloverState[localIndex] := _InZoneFillingBuffer
      rolloverCount[localIndex] := 0
    elseif rolloverState[localIndex] == _InZoneFillingBuffer
      if ++rolloverCount[localIndex] == _SamplesToAverage
        rolloverState[localIndex] := _InZoneUseBuffer
    elseif rolloverState[localIndex] == _InZoneUseBuffer
      average[localIndex] := (rolloverTotal[localIndex]  / _SamplesToAverage) // _TwelveBitsPlusOne
    if rolloverStateM[localIndex] == _OutOfZoneEmptyBuffer
      rolloverStateM[localIndex] := _InZoneFillingBuffer
      rolloverCountM[localIndex] := 0
    elseif rolloverStateM[localIndex] == _InZoneFillingBuffer
      if ++rolloverCountM[localIndex] == _SamplesToAverage
        rolloverStateM[localIndex] := _InZoneUseBuffer
    elseif rolloverStateM[localIndex] == _InZoneUseBuffer
      averageM[localIndex] := (rolloverTotalM[localIndex]  / _SamplesToAverageMoving) // _TwelveBitsPlusOne  
  else ' assume rollover problems have cleared out of buffer by now
    rolloverState[localIndex] := _OutOfZoneEmptyBuffer 
    rolloverStateM[localIndex] := _OutOfZoneEmptyBuffer

  localMovingFlag := _NotMovingThisLoop
   
  if encoderAngle[localIndex] < _MovingTrigger and previousAngle[localIndex] > _TwelveBits - _MovingTrigger
    if (encoderAngle[localIndex] + _TwelveBits) - previousAngle[localIndex] => _MovingTrigger
      localMovingFlag := _MovingThisLoop
  elseif encoderAngle[localIndex] > _TwelveBits - _MovingTrigger and previousAngle[localIndex] < _MovingTrigger
    if (previousAngle[localIndex] + _TwelveBits) -  encoderAngle[localIndex] => _MovingTrigger
      localMovingFlag := _MovingThisLoop
  elseif ||(encoderAngle[localIndex] - previousAngle[localIndex]) => _MovingTrigger  
    localMovingFlag := _MovingThisLoop

  if localMovingFlag == _NotMovingThisLoop
    if movingState[localIndex] == _Moving
      movingState[localIndex] := _NotMovingFillingBuffer
      notMovingCount[localIndex] := 1
      average[localIndex] := averageM[localIndex]
    elseif movingState[localIndex] == _NotMovingFillingBuffer
      if ++notMovingCount[localIndex] => _SamplesToAverage
        movingState[localIndex] := _NotMovingUpdatedBuffer
      else
        average[localIndex] := averageM[localIndex]
  else
    movingState[localIndex] := _Moving
    average[localIndex] := averageM[localIndex]
  
  encoderAngle7[localIndex] := average[localIndex] 
  if average[localIndex] <> previousAngle[localIndex]   '' Comment out this line to display encoder value in each loop.
    if displayMode == _ScrollDisplay 
      Debug.str(string(13, " encoder["))
      Debug.dec(localIndex)
      Debug.str(string("] = "))
      Debug.dec(average[localIndex])
    else        
      DisplayData(localIndex)
    {SevenSegment[localIndex].Char("v")
    if encoderAngle7[localIndex] < 1000
      SevenSegment[localIndex].Char("0")
    if encoderAngle7[localIndex] < 100
      SevenSegment[localIndex].Char("0") 
    if encoderAngle7[localIndex] < 10
      SevenSegment[localIndex].Char("0") 
    SevenSegment[localIndex].Dec(encoderAngle7[localIndex])
    }
    previousAngle[localIndex] := average[localIndex]
        
PUB DisplayHeading

  Debug.Clear
  Debug.Home
  Debug.str(string(13, 13, "-------------------------------------------------"))
  Debug.str(string(13, "Magnetic Encoder AS5055 Demo Program "))
  Debug.str(string(13, "The file name is "))
  Debug.str(@fileName)
  Debug.str(string(13, "The file name of the AS5055 driver is "))
  Debug.str(Spi.GetFileName)
  Debug.str(string(13, "The SPI driver was launched in cog #"))
  Debug.dec(spiCog)
  Debug.str(string(13, "-------------------------------------------------", 13))
  waitcnt(clkfreq + cnt)
  DisplayTitles
  'Debug.ClearBelow
                        
PUB DisplayTitles | localIndex

  repeat localIndex from 0 to _ChainedDevices - 1
    Debug.Position(titleX[localIndex], titleY[localIndex])
    Debug.str(@columnHeading)
    Debug.Dec(localIndex)
    
PUB DisplayData(localIndex) 

  Debug.Position(displayX[localIndex], displayY[localIndex])
  if average[localIndex] < 1000  ' space pad
    Debug.Char(32)
  if average[localIndex] < 100  
    Debug.Char(32)
  if average[localIndex] < 10  
    Debug.Char(32)  
  Debug.Dec(average[localIndex])
 
DAT

columnHeading           byte "Device #", 0